%---------------------------shared_components.m---------------------------%
%
% This file evaluates the performance of flexibly coupled CP (CCP ALS) 
% against uncoupled CP (ALS) for coupling through similar shared components
% not the entire factors. The tensors are measured with i.i.d. additive 
% Gaussian noise.
%
% These simulations are related to Fig. 2 of the paper
% "Exploring multimodal data fusion through joint decompositions with
% flexible couplings".
%
%-------------------------------------------------------------------------%

%--------------Cleaning the workspace and loading functions---------------%
clc;
clear     all;
close     all;
% Add path
addpath('CCP_ALS');
%-------------------------------------------------------------------------%

%-----------------------------Model parameters----------------------------%
% 3-way block dimensions
dim       =     [10 10 10];
% Number of components
R         =     2;
% Standard deviations of the measurements noise
sigma_n   =     [0.05 0.05];
% Standard deviation of the coupling
sigma_c   =     0.001;
% Number of iterations for coupled ALS
iter_c    =     1000;
% Number of iterations for uncoupled ALS
iter_u    =     2000;
% Number of realizations for the evaluation of the sum of the MSE
n_real    =     200;
% Number of realizations for the initializations
n_init    =     3;
%-------------------------------------------------------------------------%

%------------------------------Tensor generation--------------------------%
% Fixing the seed to have the same tensor
rng(15001)
% Random factors generation
A         =     randn(dim(1),R);
B         =     randn(dim(2),R);
A_p       =     randn(dim(1),R);
B_p       =     randn(dim(2),R);
C_p       =     randn(dim(3),R);
% Shared factor
C         =     randn(dim(3),R);
C(:,1)    =     C_p(:,1)+sigma_c*randn(dim(3),1);

% Normalization
A         =     A.*repmat(1./sqrt(sum(A.^2)),dim(1),1);
B         =     B.*repmat(1./sqrt(sum(B.^2)),dim(2),1);
A_p       =     A_p.*repmat(1./sqrt(sum(A_p.^2)),dim(1),1);
B_p       =     B_p.*repmat(1./sqrt(sum(B_p.^2)),dim(2),1);
%-------------------------------------------------------------------------%

%---------------------------------Coupling model--------------------------%
% Constraint matrices for vectorized factors
% First tensor
H   =     [eye(dim(3)) zeros([dim(3) dim(3)*(R-1)])];
% Second tensor
H_p =     [eye(dim(3)) zeros([dim(3) dim(3)*(R-1)])];
%-------------------------------------------------------------------------%


%------------------------Different noise realizations---------------------%
% Progress
N         =     n_real*n_init;
n         =     0;
% Change the seed to have different noise realizations
%rng('shuffle')

% Allocation for the squared errors
se_A_u    =     zeros(length(A(:)),n_real);
se_A_c    =     zeros(length(A(:)),n_real);
se_A_p_u  =     zeros(length(A_p(:)),n_real);
se_A_p_c  =     zeros(length(A_p(:)),n_real);
se_B_u    =     zeros(length(B(:)),n_real);
se_B_c    =     zeros(length(B(:)),n_real);
se_B_p_u  =     zeros(length(B_p(:)),n_real);
se_B_p_c  =     zeros(length(B_p(:)),n_real);
se_C_u    =     zeros(length(C(:)),n_real);
se_C_c    =     zeros(length(C(:)),n_real);
se_C_p_u  =     zeros(length(C_p(:)),n_real);
se_C_p_c  =     zeros(length(C_p(:)),n_real);

for j     =     1:n_real
   
% Noisy measurements
% First noisy tensor
Y         =     reshape(A*transpose(kr(C,B)),dim)+...
                sigma_n(1)*randn(dim);
% Second noisy tensor
Y_p       =     reshape(A_p*transpose(kr(C_p,B_p)),dim)+...
                sigma_n(2)*randn(dim);

% Different initializations
% Initial cost function values for picking best initialization
obj_n_val_1     =     10^50;
obj_n_val_2     =     10^50;
map_n_val =     10^50;
for i     =     1:n_init
%-----------------------------Initialization------------------------------%
A_0         =     randn([dim(1) R]);
B_0         =     randn([dim(2) R]);
A_0         =     A_0.*repmat(1./sqrt(sum(A_0.^2)),dim(1),1);
B_0         =     B_0.*repmat(1./sqrt(sum(B_0.^2)),dim(2),1);
   
A_p_0       =     randn([dim(1) R]);
B_p_0       =     randn([dim(2) R]);
A_p_0       =     A_p_0.*repmat(1./sqrt(sum(A_p_0.^2)),dim(1),1);
B_p_0       =     B_p_0.*repmat(1./sqrt(sum(B_p_0.^2)),dim(2),1);
 
C_p_0       =     randn([dim(3) R]);
C_0         =     randn([dim(3) R]);
%-------------------------------------------------------------------------%

%-----------------------------Uncoupled CP ALS----------------------------%
% ALS for the first tensor
[A_n_u,B_n_u,C_n_u,obj_val_vec_1] =     als(Y,iter_u,A_0,B_0,C_0);
% ALS for the second tensor
[A_p_n_u,B_p_n_u,C_p_n_u,obj_val_vec_2] = ...
                                         als(Y_p,iter_u,A_p_0,B_p_0,C_p_0);
[A_p_n_u,B_p_n_u,C_p_n_u,A_n_u,B_n_u,C_n_u]               =    ...
                             amb_correct_1f(A_p_n_u,B_p_n_u,C_p_n_u,A_n_u,B_n_u,C_n_u);
if obj_val_vec_1(end)<obj_n_val_1
obj_n_val_1     =     obj_val_vec_1(end);
A_u             =     A_n_u;
B_u             =     B_n_u;
C_u             =     C_n_u;
end
if obj_val_vec_2(end)<obj_n_val_2
obj_n_val_2     =     obj_val_vec_2(end);
A_p_u           =     A_p_n_u;
B_p_u           =     B_p_n_u;
C_p_u           =     C_p_n_u;
end
%-------------------------------------------------------------------------%

%-----------------------Flexibly coupled CP ALS---------------------------%
A_0       =     A_n_u;
B_0       =     B_n_u;
C_0       =     C_n_u;
A_p_0     =     A_p_n_u;
B_p_0     =     B_p_n_u;
C_p_0     =     C_p_n_u;
% CCP_ALS
[A_n_c,B_n_c,C_n_c,A_p_n_c,B_p_n_c,C_p_n_c,map_val_vec] = ...
                     ccp_als(Y,Y_p,iter_c,H,H_p,sigma_n,sigma_c,'joint',...
                                            A_0,B_0,C_0,A_p_0,B_p_0,C_p_0);

if map_val_vec(end)<map_n_val
map_n_val =     map_val_vec(end);
A_c       =     A_n_c;
B_c       =     B_n_c;
C_c       =     C_n_c;

A_p_c     =     A_p_n_c;
B_p_c     =     B_p_n_c;
C_p_c     =     C_p_n_c;
end
%-------------------------------------------------------------------------%

% Progress
n=n+1;
clc
fprintf('Progress %g %%',(n/N)*100)
end
%-------------------------------------------------------------------------%

%--------Correcting the ambiguity and evaluating the squared error--------%
% Uncoupled CP ALS
% Correcting permutation and sign ambiguity for the first tensor
[A_u,B_u,C_u]         =     amb_correct(A_u,B_u,C_u,A,B,C);

% Correcting permutation and sign ambiguity for the second tensor
[A_p_u,B_p_u,C_p_u]   =...
                      amb_correct(A_p_u,B_p_u,C_p_u,A_p,B_p,C_p);
% Squared errors on the factors
se_A_u(:,j)     =     (A(:)-A_u(:)).^2;
se_B_u(:,j)     =     (B(:)-B_u(:)).^2;
se_C_u(:,j)     =     (C(:)-C_u(:)).^2;
se_A_p_u(:,j)   =     (A_p(:)-A_p_u(:)).^2;
se_B_p_u(:,j)   =     (B_p(:)-B_p_u(:)).^2;
se_C_p_u(:,j)   =     (C_p(:)-C_p_u(:)).^2;

% Flexibly coupled CP ALS
% Correcting permutation and sign ambiguity for the first tensor
[A_c,B_c,C_c]   =     amb_correct(A_c,B_c,C_c,A,B,C);

% Correcting permutation and sign ambiguity for the second tensor
[A_p_c,B_p_c,C_p_c]   =     amb_correct(A_p_c,...
                                        B_p_c,C_p_c,A_p,B_p,C_p);
% Squared errors
se_A_c(:,j)     =     (A(:)-A_c(:)).^2;
se_B_c(:,j)     =     (B(:)-B_c(:)).^2;
se_C_c(:,j)     =     (C(:)-C_c(:)).^2;
se_A_p_c(:,j)   =     (A_p(:)-A_p_c(:)).^2;
se_B_p_c(:,j)   =     (B_p(:)-B_p_c(:)).^2;
se_C_p_c(:,j)   =     (C_p(:)-C_p_c(:)).^2;
%-------------------------------------------------------------------------%
end
%-------------------------------------------------------------------------%

%---------------------------Results Fig. 2--------------------------------%
% Factor C
plot(1:20,mean(se_C_u,2),'b*',1:20,...
     mean(se_C_c,2),'k+')
ylabel('Total MSE on C')
xlabel('Component')
legend('Uncoupled','Coupled')
 
% Factor C_p
figure
plot(1:20,mean(se_C_p_u,2),'b*',1:20,...
     mean(se_C_p_c,2),'k+')
ylabel('Total MSE on C_p')
xlabel('Component')
legend('Uncoupled','Coupled')
%-------------------------------------------------------------------------%

