%----------------------------interpolation_1.m----------------------------%
%
% This file evaluates the performance of coupled CP (CCP ALS) with interpo-
% lation coupling against uncoupled CP (ALS) for different noise levels
% for the first tensor (the noise level on the other tensor is fixed). The 
% tensors are measured with i.i.d. additive Gaussian noise.
%
% The vectors in the factor C are sampled versions of a sum of three sine
% signals (bandlimited periodic signal).
%
% These simulations are related to Fig. 4 and Fig. 5 of the paper
% "Exploring multimodal data fusion through joint decompositions with
% flexible couplings".
%
%-------------------------------------------------------------------------%

%--------------Cleaning the workspace and loading functions---------------%
clc;
clear     all;
close     all;
% Add path
addpath('CCP_ALS');
%-------------------------------------------------------------------------%

%-----------------------------Model parameters----------------------------%
% 3-way block dimensions
dim       =     [10 10 37];
% Second tensor
dim_p     =     [10 10 53];
% Oversampling dimension
dim_os    =     100;
% Time duration
T_dur     =     4;
% Number of components
R         =     3;
% Sine frequencies
f_s_1     =     2.05;
f_s_2     =     2.55;
f_s_3     =     3.5;
% Standard deviations of the measurements noise
sigma_n   =     [0.001 0.01; 0.002 0.01; 0.005 0.01; 0.01 0.01;0.02 0.01...
                 ; 0.05 0.01; 0.1 0.01; 0.2 0.01];
% Standard deviation of the coupling
sigma_c   =     0.15;
%sigma_c   =    0.0001;% Approx. hard coupling
% Number of iterations for the CCP ALS
iter_c    =     200;
% Number of iterations for the uncoupled ALS
iter_u    =     1000;
% Number of realizations for the evaluation of the sum of the MSE
n_real    =     200;
% Number of realizations for the initializations
n_init    =     10;
%-------------------------------------------------------------------------%

%------------------------------Tensor generation--------------------------%
% Fixing the seed to have the same tensor
rng(102)
% Random factors generation
A         =     randn(dim(1),R);
B         =     randn(dim(2),R);
A_p       =     randn(dim_p(1),R);
B_p       =     randn(dim_p(2),R);
% Normalization
A         =     A.*repmat(1./sqrt(sum(A.^2)),dim(1),1);
B         =     B.*repmat(1./sqrt(sum(B.^2)),dim(2),1);
A_p       =     A_p.*repmat(1./sqrt(sum(A_p.^2)),dim_p(1),1);
B_p       =     B_p.*repmat(1./sqrt(sum(B_p.^2)),dim_p(2),1);

% Factors C coupled by sampling
% Sine amplitudes
c_1_1     =     randn(1,1);
c_1_2     =     randn(1,1);
c_1_3     =     randn(1,1);
c_2_1     =     randn(1,1);
c_2_2     =     randn(1,1);
c_2_3     =     randn(1,1);
c_3_1     =     randn(1,1);
c_3_2     =     randn(1,1);
c_3_3     =     randn(1,1);

% Sampling frequencies
% First matrix
T_s       =     T_dur/(dim(3));
% Second matrix
T_s_p     =     T_dur/(dim_p(3));

% First C factor
k         =     0:dim(3)-1;
t         =     (k*T_s)';
C         =     [c_1_1*sin(2*pi*f_s_1*t)+...
                 c_1_2*sin(2*pi*f_s_2*t)+...
                 c_1_3*sin(2*pi*f_s_3*t) ...
                 c_2_1*sin(2*pi*f_s_1*t)+...
                 c_2_2*sin(2*pi*f_s_2*t)+...
                 c_2_3*sin(2*pi*f_s_3*t) ...
                 c_3_1*sin(2*pi*f_s_1*t)+...
                 c_3_2*sin(2*pi*f_s_2*t)+...
                 c_3_3*sin(2*pi*f_s_3*t)];%+...
                                     %(sigma_12/sqrt(2)*randn(R,dim(2)));
% Second C factor
k_p       =     0:dim_p(3)-1;
t_p       =     (k_p*T_s_p)';
C_p       =     [c_1_1*sin(2*pi*f_s_1*t_p)+...
                 c_1_2*sin(2*pi*f_s_2*t_p)+...
                 c_1_3*sin(2*pi*f_s_3*t_p) ...
                 c_2_1*sin(2*pi*f_s_1*t_p)+...
                 c_2_2*sin(2*pi*f_s_2*t_p)+...
                 c_2_3*sin(2*pi*f_s_3*t_p) ...
                 c_3_1*sin(2*pi*f_s_1*t_p)+...
                 c_3_2*sin(2*pi*f_s_2*t_p)+...
                 c_3_3*sin(2*pi*f_s_3*t_p)];%+...
                                     %(sigma_12/sqrt(2)*randn(R,dim_p(2)));
%-------------------------------------------------------------------------%


%---------------------------------Coupling model--------------------------%
% Oversampled period
T_os      =     T_dur/(dim_os);
% Oversampled grid
k_os      =     0:(dim_os-1);
t_os      =     (k_os*T_os)';
% Interpolating kernel matrices
% First tensor
H_mat     =     zeros(dim_os,dim(3));
for l     =     1:dim_os
    for   k     =     1:dim(3)
          arg_sin     =     (t_os(l)-t(k))/(T_dur);
          if mod(arg_sin,1) == 0
          H_mat(l,k)  =     1;
          else
          H_mat(l,k)  =     sin(dim(3)*pi*arg_sin)/...
                            (dim(3)*sin(pi*arg_sin));
          end
    end
end
% Second tensor
H_p_mat   =     zeros(dim_os,dim_p(3));
for l     =     1:dim_os
    for   k     =     1:dim_p(3)
          arg_sin     =     (t_os(l)-t_p(k))/(T_dur);
          if mod(arg_sin,1) == 0
          H_p_mat(l,k)=     1;
          else
          H_p_mat(l,k)=     sin(dim_p(3)*pi*arg_sin)/...
                            (dim_p(3)*sin(pi*arg_sin));
          end
    end
end
% Constraint matrices for vectorized factors
% First tensor
H   =     kron(eye(R),H_mat);
% Second tensor
H_p =     kron(eye(R),H_p_mat);

% Displaying the interpolation
C_os      =     H_mat*(C);
C_p_os    =     H_p_mat*(C_p);

% Continuous underlying functions
dim_c     =     200;
t_c       =     linspace(0,4,dim_c)';
C_c       =     [c_1_1*sin(2*pi*f_s_1*t_c)+...
                 c_1_2*sin(2*pi*f_s_2*t_c)+...
                 c_1_3*sin(2*pi*f_s_3*t_c) ...
                 c_2_1*sin(2*pi*f_s_1*t_c)+...
                 c_2_2*sin(2*pi*f_s_2*t_c)+...
                 c_2_3*sin(2*pi*f_s_3*t_c) ...
                 c_3_1*sin(2*pi*f_s_1*t_c)+...
                 c_3_2*sin(2*pi*f_s_2*t_c)+...
                 c_3_3*sin(2*pi*f_s_3*t_c)];
%-------------------------------------------------------------------------%

%----------------------------Results Fig. 4-------------------------------%
% Different samplings
plot(t_c,C_c(:,1),'k',t,C(:,1),'b*',t_p,C_p(:,1),'r+')
legend('Continuous signal','First sampling','Second sampling')
%-------------------------------------------------------------------------%

%---------------Simulation for different noise levels---------------------%
% Progress
N         =     n_real*n_init*size(sigma_n,1);
n         =     0;

% Allocation for the squared errors
se_A_u    =     zeros(length(A(:)),n_real,size(sigma_n,1));
se_A_c    =     zeros(length(A(:)),n_real,size(sigma_n,1));
se_A_p_u  =     zeros(length(A_p(:)),n_real,size(sigma_n,1));
se_A_p_c  =     zeros(length(A_p(:)),n_real,size(sigma_n,1));
se_B_u    =     zeros(length(B(:)),n_real,size(sigma_n,1));
se_B_c    =     zeros(length(B(:)),n_real,size(sigma_n,1));
se_B_p_u  =     zeros(length(B_p(:)),n_real,size(sigma_n,1));
se_B_p_c  =     zeros(length(B_p(:)),n_real,size(sigma_n,1));
se_C_u    =     zeros(length(C(:)),n_real,size(sigma_n,1));
se_C_c    =     zeros(length(C(:)),n_real,size(sigma_n,1));
se_C_p_u  =     zeros(length(C_p(:)),n_real,size(sigma_n,1));
se_C_p_c  =     zeros(length(C_p(:)),n_real,size(sigma_n,1));

% Uncoupled ALS
mse_A_u   =     zeros(1,size(sigma_n,1));
mse_B_u   =     zeros(1,size(sigma_n,1));
mse_C_u   =     zeros(1,size(sigma_n,1));
mse_A_p_u =     zeros(1,size(sigma_n,1));
mse_B_p_u =     zeros(1,size(sigma_n,1));
mse_C_p_u =     zeros(1,size(sigma_n,1));
% Total MSE on the C factors 
t_err_u   =     zeros(1,size(sigma_n,1));

% CCP ALS
mse_A_c   =     zeros(1,size(sigma_n,1));
mse_B_c   =     zeros(1,size(sigma_n,1));
mse_C_c   =     zeros(1,size(sigma_n,1));
mse_A_p_c =     zeros(1,size(sigma_n,1));
mse_B_p_c =     zeros(1,size(sigma_n,1));
mse_C_p_c =     zeros(1,size(sigma_n,1));
% Total MSE on the C factors
t_err_c   =     zeros(1,size(sigma_n,1));

% Change the seed to have different noise realizations
%rng('shuffle')
for k     =     1:size(sigma_n,1)
%------------------------Different noise realizations---------------------%
for j     =     1:n_real
%Noisy measurements
% First noisy tensor
Y         =     reshape(A*transpose(kr(C,B)),dim)+...
                sigma_n(k,1)*randn(dim);
% Second noisy tensor
Y_p       =     reshape(A_p*transpose(kr(C_p,B_p)),dim_p)+...
                sigma_n(k,2)*randn(dim_p);

% Different initializations
% Initial cost function values for picking best initialization
obj_n_val_1     =     10^50;
obj_n_val_2     =     10^50;
map_n_val       =     10^50;
for i     =     1:n_init
%-----------------------------Initialization------------------------------%
% Initialization around the true factors
A_0         =     randn([dim(1) R]);
B_0         =     randn([dim(2) R]);
A_0         =     A_0.*repmat(1./sqrt(sum(A_0.^2)),dim(1),1);
B_0         =     B_0.*repmat(1./sqrt(sum(B_0.^2)),dim(2),1);
   
A_p_0       =     randn([dim_p(1) R]);
B_p_0       =     randn([dim_p(2) R]);
A_p_0       =     A_p_0.*repmat(1./sqrt(sum(A_p_0.^2)),dim_p(1),1);
B_p_0       =     B_p_0.*repmat(1./sqrt(sum(B_p_0.^2)),dim_p(2),1);
 
C_p_0       =     randn([dim_p(3) R]);
C_0         =     randn([dim(3) R]);
%-------------------------------------------------------------------------%

%---------------------------Uncoupled CP ALS------------------------------%
% ALS for the first tensor
[A_n_u,B_n_u,C_n_u,obj_val_vec_1] =     als(Y,iter_u,A_0,B_0,C_0);
% ALS for the second tensor
[A_p_n_u,B_p_n_u,C_p_n_u,obj_val_vec_2] = ...
                                         als(Y_p,iter_u,A_p_0,B_p_0,C_p_0);
[A_p_n_u,B_p_n_u,C_p_n_u]               =     ...
                       amb_correct_transf(A_p_n_u,B_p_n_u,C_p_n_u,C_n_u,...
                                                            H_mat,H_p_mat);
if obj_val_vec_1(end)<obj_n_val_1
obj_n_val_1     =     obj_val_vec_1(end);
A_u       =     A_n_u;
B_u       =     B_n_u;
C_u       =     C_n_u;
end
if obj_val_vec_2(end)<obj_n_val_2
obj_n_val_2     =     obj_val_vec_2(end);
A_p_u     =     A_p_n_u;
B_p_u     =     B_p_n_u;
C_p_u     =     C_p_n_u;
end
%-------------------------------------------------------------------------%

%---------------------Flexibly coupled CP ALS-----------------------------%
A_0       =     A_n_u;
B_0       =     B_n_u;
C_0       =     C_n_u;
A_p_0     =     A_p_n_u;
B_p_0     =     B_p_n_u;
C_p_0     =     C_p_n_u;
% CCP_ALS
[A_n_c,B_n_c,C_n_c,A_p_n_c,B_p_n_c,C_p_n_c,map_val_vec] = ...
    ccp_als(Y,Y_p,iter_c,H,H_p,sigma_n(k,:),sigma_c,'joint',A_0,B_0,C_0,...
                                                        A_p_0,B_p_0,C_p_0);

if map_val_vec(end)<map_n_val
map_n_val       =     map_val_vec(end);
A_c       =     A_n_c;
B_c       =     B_n_c;
C_c       =     C_n_c;

A_p_c     =     A_p_n_c;
B_p_c     =     B_p_n_c;
C_p_c     =     C_p_n_c;
end
%-------------------------------------------------------------------------%
% Progress
n=n+1;
clc
fprintf('Progress %g %%',(n/N)*100)
end

%--------Correcting the ambiguity and evaluating the squared error--------%
% Uncoupled CP ALS
% Correcting permutation and sign ambiguity for the first tensor
[A_u,B_u,C_u]      =     amb_correct(A_u,B_u,C_u,A,B,C);
% Correcting permutation and sign ambiguity for the second tensor
[A_p_u,B_p_u,C_p_u]=...
                       amb_correct(A_p_u,B_p_u,C_p_u,A_p,B_p,C_p);
% Squared errors on the factors
se_A_u(:,j,k)   =     (A(:)-A_u(:)).^2;
se_B_u(:,j,k)   =     (B(:)-B_u(:)).^2;
se_C_u(:,j,k)   =     (C(:)-C_u(:)).^2;
se_A_p_u(:,j,k) =     (A_p(:)-A_p_u(:)).^2;
se_B_p_u(:,j,k) =     (B_p(:)-B_p_u(:)).^2;
se_C_p_u(:,j,k) =     (C_p(:)-C_p_u(:)).^2;

% Flexibly coupled CP ALS
% Correcting permutation and sign ambiguity for the first tensor
[A_c,B_c,C_c]   =     amb_correct(A_c,B_c,C_c,A,B,C);
% Correcting permutation and sign ambiguity for the second tensor
[A_p_c,B_p_c,C_p_c]   =     amb_correct(A_p_c,...
                                        B_p_c,C_p_c,A_p,B_p,C_p);
% Squared errors
se_A_c(:,j,k)         =     (A(:)-A_c(:)).^2;
se_B_c(:,j,k)         =     (B(:)-B_c(:)).^2;
se_C_c(:,j,k)         =     (C(:)-C_c(:)).^2;
se_A_p_c(:,j,k)       =     (A_p(:)-A_p_c(:)).^2;
se_B_p_c(:,j,k)       =     (B_p(:)-B_p_c(:)).^2;
se_C_p_c(:,j,k)       =     (C_p(:)-C_p_c(:)).^2;
%-------------------------------------------------------------------------%
end
%-------------------------------------------------------------------------%

%-----------------------------Results-------------------------------------%
% Mean squared errors on the factors
% Uncoupled CP ALS
mse_A_u(k)      =     sum(mean(se_A_u(:,:,k),2));
mse_B_u(k)      =     sum(mean(se_B_u(:,:,k),2));
mse_C_u(k)      =     sum(mean(se_C_u(:,:,k),2));
mse_A_p_u(k)    =     sum(mean(se_A_p_u(:,:,k),2));
mse_B_p_u(k)    =     sum(mean(se_B_p_u(:,:,k),2));
mse_C_p_u(k)    =     sum(mean(se_C_p_u(:,:,k),2));
% Total MSE on the C factors 
t_err_u(k)      =     mse_C_u(k)+mse_C_p_u(k);
% Flexibly coupled CP ALS
mse_A_c(k)      =     sum(mean(se_A_c(:,:,k),2));
mse_B_c(k)      =     sum(mean(se_B_c(:,:,k),2));
mse_C_c(k)      =     sum(mean(se_C_c(:,:,k),2));
mse_A_p_c(k)    =     sum(mean(se_A_p_c(:,:,k),2));
mse_B_p_c(k)    =     sum(mean(se_B_p_c(:,:,k),2));
mse_C_p_c(k)    =     sum(mean(se_C_p_c(:,:,k),2));
% Total MSE on the C factors
t_err_c(k)      =     mse_C_c(k)+mse_C_p_c(k);
end
%-------------------------------------------------------------------------%

%-----------------------------Results Fig. 5------------------------------%
SNR_diff=20*log10(sigma_n(:,1)/sigma_n(1,2));
semilogy(SNR_diff,t_err_u,'r',SNR_diff,t_err_c,'b')
xlabel('SNR difference')
ylabel('Total MSE on C and C_p')
legend('Uncoupled','Coupled')
%-------------------------------------------------------------------------%